import asyncio
import logging
import os
from pathlib import Path

from aiohttp_sse import sse_response
from aiohttp.web import Response, FileResponse, HTTPFound, HTTPNotFound
from jinja2 import Environment, PackageLoader, select_autoescape
from multidict import MultiDict

from .app_config import extract_app_config, render_app_html
from .form import build_form, ALLOWED_FILE_TYPES
from .samples import TWITTER_FEED, INSTAGRAM_FEED
from .storage import get_file, save_file
from .utils import (
    formdata_to_json,
    inject_script_into_html,
)


jinja_env = Environment(
    loader=PackageLoader("app_simulator", "templates"),
    autoescape=select_autoescape(["html", "xml"]),
)

logger = logging.getLogger("onsigntv.routes")


tracked_files = {}


def track_file(x):
    tracked_files[x] = 0


async def list_form_file(request):
    if "file_name" in request.match_info:
        name = request.match_info["file_name"]

        if name == "..back":
            name = ".."
    else:
        name = ""

    path = Path.cwd() / name

    if path.is_file() and path.name.endswith(".html"):
        logger.debug(f"open a html file: {name}")

        try:
            config = extract_app_config(path.read_text("utf-8"))

            tracked_files.clear()
            logger.debug(f"add {name} to tracked files list")
            track_file(path)

            form = build_form(config)
            form.process()

        except Exception as exp:
            exceptions = [exp]
            logger.info(f"handling exception: {exp}")

            return Response(
                text=jinja_env.get_template("widget_exceptions.html").render(
                    {"exceptions": exceptions}
                ),
                content_type="text/html",
            )

        return Response(
            text=jinja_env.get_template("widget_form.html").render(
                {
                    "form": form,
                    "title": config["title"],
                    "file_name": "/.preview/{}".format(name),
                },
            ),
            content_type="text/html",
        )
    elif path.is_file():
        if path.suffix in ALLOWED_FILE_TYPES:
            logger.debug(f"serving file: {name}")
            return FileResponse(name)
        else:
            return HTTPNotFound()
    else:
        logger.debug("opening directory")
        if path.is_dir() and name != "":
            logger.debug("back to a parent directory")
            os.chdir(name)
            raise HTTPFound("/")

        folder_or_files = [{"name": "..", "link": "/..back"}]
        for element in sorted(Path.cwd().iterdir(), key=lambda f: f.name.lower()):
            if not element.name.startswith("."):
                folder_or_files.append(
                    {
                        "name": element.name,
                        "link": "/" + element.name,
                        "is_file": Path.is_file(Path.cwd() / element),
                    }
                )

        return Response(
            text=jinja_env.get_template("list_files.html").render(
                {"files": folder_or_files}
            ),
            content_type="text/html",
        )


async def preview_app(request):
    logger.debug("retrieve data")

    name = request.match_info["file_name"]
    path = Path.cwd() / name

    if path not in tracked_files:
        logger.debug(f"add {name} to tracked files list before preview")
        track_file(path)

    formdata = MultiDict()

    request_reader = await request.multipart()

    while True:
        part = await request_reader.next()
        if part is None:
            break

        field_name = part.name
        if part.filename:
            logger.debug(f"field name {field_name} is a file")
            file_byte = await part.read(decode=True)
            formdata.add(field_name, save_file(part.filename, file_byte))
        else:
            metadata = await part.text()
            logger.debug(f"field name {field_name} has value {metadata}")
            formdata.add(field_name, metadata)

    try:
        config = extract_app_config(path.read_text("utf-8"))
    except Exception as exp:
        exceptions = [exp]
        logger.debug(f"handling exception: {exp}")
        logger.info(f"handling exception: {exp}")

        return Response(
            text=jinja_env.get_template("widget_exceptions.html").render(
                {
                    "exceptions": exceptions,
                    "formdata": formdata_to_json(formdata),
                },
            ),
            content_type="text/html",
        )

    form = build_form(config)
    form.process(formdata=formdata)

    if form.validate():
        logger.debug("form data valid")
        data = {}
        for field in form:
            if callable(getattr(field, "adapt", None)):
                field.adapt()

            data[field.name] = field.data

        html, excep = render_app_html(data, path, track_file)
        if html is None:
            logger.debug("handling rendering exceptions")
            return Response(
                text=jinja_env.get_template("widget_exceptions.html").render(
                    {
                        "exceptions": [str(excep)],
                        "formdata": formdata_to_json(formdata),
                    },
                ),
                content_type="text/html",
            )

        html = inject_script_into_html(html, formdata)

        return Response(text=html, content_type="text/html")
    else:
        logger.debug("invalid data")

        for field in form:
            for error in field.errors:
                logger.info(f"{field.name} : {error}")

        return Response(
            text=jinja_env.get_template("widget_form.html").render(
                {
                    "form": form,
                    "title": config["title"],
                    "file_name": "/.preview/{}".format(request.match_info["file_name"]),
                },
            ),
            content_type="text/html",
        )


async def change_notification_sse(request):
    loop = request.app.loop
    async with sse_response(request) as resp:
        while True:
            if not tracked_files:
                await resp.send("refresh")
                return resp

            for path, last_modification in tracked_files.items():
                if path.is_file():
                    actual_modification = path.stat().st_ctime
                    logger.debug(f"Checking m_time {path}")

                    if last_modification == 0:
                        tracked_files[path] = actual_modification
                    else:
                        if actual_modification > last_modification:
                            logger.info(f"Reloading due to change on {path}")
                            await resp.send("refresh")
                            tracked_files[path] = actual_modification
                else:
                    logger.info(f"File not found: {path}")
            await asyncio.sleep(1, loop=loop)

    return resp


async def serve_file_from_uploads(request):
    return FileResponse(get_file(request.match_info["file_name"]))


async def serve_twitter_data(request):
    from aiohttp import web
    import json

    return web.json_response(json.loads(TWITTER_FEED))


async def serve_instagram_data(request):
    from aiohttp import web
    import json

    return web.json_response(json.loads(INSTAGRAM_FEED))



INSTAGRAM_FEED = """
[
    {
        "comments": 2,
        "likes": 1,
        "id": "363839373298",
        "created": "2010-07-17 01:29:43",
        "caption": "",
        "type": "video",
        "tags": [
        ],
        "user": {
            "username": "kevin",
            "full_name": "Kevin S",
            "profile_picture": "http://distillery.s3.amazonaws.com/profiles/profile_3_75sq_1295574122.jpg",
            "id": "3"
        },
        "image": {
            "url": "http://distilleryimage2.ak.instagram.com/11f75f1cd9cc11e2a0fd22000aa8039a_7.jpg",
            "width": 612,
            "height": 612
        }
    },
    {
        "comments": 0,
        "likes": 15,
        "id": "26621408",
        "created": "2011-02-03 03:19:12",
        "caption": "Inside le truc #foodtruck",
        "type": "image",
        "tags": [
            "foodtruck"
        ],
        "user": {
            "username": "kevin",
            "full_name": "Kevin S",
            "profile_picture": "http://distillery.s3.amazonaws.com/profiles/profile_3_75sq_1295574122.jpg",
            "id": "3"
        },
        "image": {
            "url": "http://distillery.s3.amazonaws.com/media/2011/02/02/6ea7baea55774c5e81e7e3e1f6e791a7_7.jpg",
            "width": 612,
            "height": 612
        }
    }
]"""

TWITTER_FEED = """
[
    {
        "created_at": "Thu Apr 06 15:28:43 +0000 2017",
        "id": 850007368138018817,
        "id_str": "850007368138018817",
        "text": "RT @TwitterDev: 1/ Today we’re sharing our vision for the future of the Twitter API platform!nhttps://t.co/XweGngmxlP",
        "truncated": false,
        "entities": {
            "hashtags": [
            ],
            "symbols": [
            ],
            "user_mentions": [
                {
                    "screen_name": "TwitterDev",
                    "name": "TwitterDev",
                    "id": 2244994945,
                    "id_str": "2244994945",
                    "indices": [
                        3,
                        14
                    ]
                }
            ],
            "urls": [
                {
                    "url": "https://t.co/XweGngmxlP",
                    "expanded_url": "https://cards.twitter.com/cards/18ce53wgo4h/3xo1c",
                    "display_url": "cards.twitter.com/cards/18ce53wg…",
                    "indices": [
                        94,
                        117
                    ]
                }
            ]
        },
        "source": "<a href='http://twitter.com' rel='nofollow'>Twitter Web Client</a>",
        "in_reply_to_status_id": null,
        "in_reply_to_status_id_str": null,
        "in_reply_to_user_id": null,
        "in_reply_to_user_id_str": null,
        "in_reply_to_screen_name": null,
        "user": {
            "id": 6253282,
            "id_str": "6253282",
            "name": "Twitter API",
            "screen_name": "twitterapi",
            "location": "San Francisco, CA",
            "description": "The Real Twitter API. I tweet about API changes, service issues and happily answer questions about Twitter and our API. Don't get an answer? It's on my website.",
            "url": "http://t.co/78pYTvWfJd",
            "entities": {
                "url": {
                    "urls": [
                        {
                            "url": "http://t.co/78pYTvWfJd",
                            "expanded_url": "https://dev.twitter.com",
                            "display_url": "dev.twitter.com",
                            "indices": [
                                0,
                                22
                            ]
                        }
                    ]
                },
                "description": {
                    "urls": [
                    ]
                }
            },
            "protected": false,
            "followers_count": 6172353,
            "friends_count": 46,
            "listed_count": 13091,
            "created_at": "Wed May 23 06:01:13 +0000 2007",
            "favourites_count": 26,
            "utc_offset": -25200,
            "time_zone": "Pacific Time (US & Canada)",
            "geo_enabled": true,
            "verified": true,
            "statuses_count": 3583,
            "lang": "en",
            "contributors_enabled": false,
            "is_translator": false,
            "is_translation_enabled": false,
            "profile_background_color": "C0DEED",
            "profile_background_image_url": "http://pbs.twimg.com/profile_background_images/656927849/miyt9dpjz77sc0w3d4vj.png",
            "profile_background_image_url_https": "https://pbs.twimg.com/profile_background_images/656927849/miyt9dpjz77sc0w3d4vj.png",
            "profile_background_tile": true,
            "profile_image_url": "http://pbs.twimg.com/profile_images/2284174872/7df3h38zabcvjylnyfe3_normal.png",
            "profile_image_url_https": "https://pbs.twimg.com/profile_images/2284174872/7df3h38zabcvjylnyfe3_normal.png",
            "profile_banner_url": "https://pbs.twimg.com/profile_banners/6253282/1431474710",
            "profile_link_color": "0084B4",
            "profile_sidebar_border_color": "C0DEED",
            "profile_sidebar_fill_color": "DDEEF6",
            "profile_text_color": "333333",
            "profile_use_background_image": true,
            "has_extended_profile": false,
            "default_profile": false,
            "default_profile_image": false,
            "following": true,
            "follow_request_sent": false,
            "notifications": false,
            "translator_type": "regular"
        },
        "geo": null,
        "coordinates": null,
        "place": null,
        "contributors": null,
        "retweeted_status": {
            "created_at": "Thu Apr 06 15:24:15 +0000 2017",
            "id": 850006245121695744,
            "id_str": "850006245121695744",
            "text": "1/ Today we’re sharing our vision for the future of the Twitter API platform!nhttps://t.co/XweGngmxlP",
            "truncated": false,
            "entities": {
                "hashtags": [
                ],
                "symbols": [
                ],
                "user_mentions": [
                ],
                "urls": [
                    {
                        "url": "https://t.co/XweGngmxlP",
                        "expanded_url": "https://cards.twitter.com/cards/18ce53wgo4h/3xo1c",
                        "display_url": "cards.twitter.com/cards/18ce53wg…",
                        "indices": [
                            78,
                            101
                        ]
                    }
                ]
            },
            "source": "<a href='http://twitter.com' rel='nofollow'>Twitter Web Client</a>",
            "in_reply_to_status_id": null,
            "in_reply_to_status_id_str": null,
            "in_reply_to_user_id": null,
            "in_reply_to_user_id_str": null,
            "in_reply_to_screen_name": null,
            "user": {
                "id": 2244994945,
                "id_str": "2244994945",
                "name": "TwitterDev",
                "screen_name": "TwitterDev",
                "location": "Internet",
                "description": "Your official source for Twitter Platform news, updates & events. Need technical help? Visit https://t.co/mGHnxZCxkt ⌨️  #TapIntoTwitter",
                "url": "https://t.co/66w26cua1O",
                "entities": {
                    "url": {
                        "urls": [
                            {
                                "url": "https://t.co/66w26cua1O",
                                "expanded_url": "https://dev.twitter.com/",
                                "display_url": "dev.twitter.com",
                                "indices": [
                                    0,
                                    23
                                ]
                            }
                        ]
                    },
                    "description": {
                        "urls": [
                            {
                                "url": "https://t.co/mGHnxZCxkt",
                                "expanded_url": "https://twittercommunity.com/",
                                "display_url": "twittercommunity.com",
                                "indices": [
                                    93,
                                    116
                                ]
                            }
                        ]
                    }
                },
                "protected": false,
                "followers_count": 465425,
                "friends_count": 1523,
                "listed_count": 1168,
                "created_at": "Sat Dec 14 04:35:55 +0000 2013",
                "favourites_count": 2098,
                "utc_offset": -25200,
                "time_zone": "Pacific Time (US & Canada)",
                "geo_enabled": true,
                "verified": true,
                "statuses_count": 3031,
                "lang": "en",
                "contributors_enabled": false,
                "is_translator": false,
                "is_translation_enabled": false,
                "profile_background_color": "FFFFFF",
                "profile_background_image_url": "http://abs.twimg.com/images/themes/theme1/bg.png",
                "profile_background_image_url_https": "https://abs.twimg.com/images/themes/theme1/bg.png",
                "profile_background_tile": false,
                "profile_image_url": "http://pbs.twimg.com/profile_images/530814764687949824/npQQVkq8_normal.png",
                "profile_image_url_https": "https://pbs.twimg.com/profile_images/530814764687949824/npQQVkq8_normal.png",
                "profile_banner_url": "https://pbs.twimg.com/profile_banners/2244994945/1396995246",
                "profile_link_color": "0084B4",
                "profile_sidebar_border_color": "FFFFFF",
                "profile_sidebar_fill_color": "DDEEF6",
                "profile_text_color": "333333",
                "profile_use_background_image": false,
                "has_extended_profile": false,
                "default_profile": false,
                "default_profile_image": false,
                "following": true,
                "follow_request_sent": false,
                "notifications": false,
                "translator_type": "regular"
            },
            "geo": null,
            "coordinates": null,
            "place": null,
            "contributors": null,
            "is_quote_status": false,
            "retweet_count": 284,
            "favorite_count": 399,
            "favorited": false,
            "retweeted": false,
            "possibly_sensitive": false,
            "lang": "en"
        },
        "is_quote_status": false,
        "retweet_count": 284,
        "favorite_count": 0,
        "favorited": false,
        "retweeted": false,
        "possibly_sensitive": false,
        "lang": "en"
    },
    {
        "created_at": "Mon Apr 03 16:09:50 +0000 2017",
        "id": 848930551989915648,
        "id_str": "848930551989915648",
        "text": "RT @TwitterMktg: Starting today, businesses can request and share locations when engaging with people in Direct Messages. https://t.co/rpYn…",
        "truncated": false,
        "entities": {
            "hashtags": [
            ],
            "symbols": [
            ],
            "user_mentions": [
                {
                    "screen_name": "TwitterMktg",
                    "name": "Twitter Marketing",
                    "id": 357750891,
                    "id_str": "357750891",
                    "indices": [
                        3,
                        15
                    ]
                }
            ],
            "urls": [
            ]
        },
        "source": "<a href='http://twitter.com' rel='nofollow'>Twitter Web Client</a>",
        "in_reply_to_status_id": null,
        "in_reply_to_status_id_str": null,
        "in_reply_to_user_id": null,
        "in_reply_to_user_id_str": null,
        "in_reply_to_screen_name": null,
        "user": {
            "id": 6253282,
            "id_str": "6253282",
            "name": "Twitter API",
            "screen_name": "twitterapi",
            "location": "San Francisco, CA",
            "description": "The Real Twitter API. I tweet about API changes, service issues and happily answer questions about Twitter and our API. Don't get an answer? It's on my website.",
            "url": "http://t.co/78pYTvWfJd",
            "entities": {
                "url": {
                    "urls": [
                        {
                            "url": "http://t.co/78pYTvWfJd",
                            "expanded_url": "https://dev.twitter.com",
                            "display_url": "dev.twitter.com",
                            "indices": [
                                0,
                                22
                            ]
                        }
                    ]
                },
                "description": {
                    "urls": [

                    ]
                }
            },
            "protected": false,
            "followers_count": 6172353,
            "friends_count": 46,
            "listed_count": 13091,
            "created_at": "Wed May 23 06:01:13 +0000 2007",
            "favourites_count": 26,
            "utc_offset": -25200,
            "time_zone": "Pacific Time (US & Canada)",
            "geo_enabled": true,
            "verified": true,
            "statuses_count": 3583,
            "lang": "en",
            "contributors_enabled": false,
            "is_translator": false,
            "is_translation_enabled": false,
            "profile_background_color": "C0DEED",
            "profile_background_image_url": "http://pbs.twimg.com/profile_background_images/656927849/miyt9dpjz77sc0w3d4vj.png",
            "profile_background_image_url_https": "https://pbs.twimg.com/profile_background_images/656927849/miyt9dpjz77sc0w3d4vj.png",
            "profile_background_tile": true,
            "profile_image_url": "http://pbs.twimg.com/profile_images/2284174872/7df3h38zabcvjylnyfe3_normal.png",
            "profile_image_url_https": "https://pbs.twimg.com/profile_images/2284174872/7df3h38zabcvjylnyfe3_normal.png",
            "profile_banner_url": "https://pbs.twimg.com/profile_banners/6253282/1431474710",
            "profile_link_color": "0084B4",
            "profile_sidebar_border_color": "C0DEED",
            "profile_sidebar_fill_color": "DDEEF6",
            "profile_text_color": "333333",
            "profile_use_background_image": true,
            "has_extended_profile": false,
            "default_profile": false,
            "default_profile_image": false,
            "following": true,
            "follow_request_sent": false,
            "notifications": false,
            "translator_type": "regular"
        },
        "geo": null,
        "coordinates": null,
        "place": null,
        "contributors": null,
        "retweeted_status": {
            "created_at": "Mon Apr 03 16:05:05 +0000 2017",
            "id": 848929357519241216,
            "id_str": "848929357519241216",
            "text": "Starting today, businesses can request and share locations when engaging with people in Direct Messages. https://t.co/rpYndqWfQw",
            "truncated": false,
            "entities": {
                "hashtags": [
                ],
                "symbols": [
                ],
                "user_mentions": [
                ],
                "urls": [
                    {
                        "url": "https://t.co/rpYndqWfQw",
                        "expanded_url": "https://cards.twitter.com/cards/5wzucr/3x700",
                        "display_url": "cards.twitter.com/cards/5wzucr/3…",
                        "indices": [
                            105,
                            128
                        ]
                    }
                ]
            },
            "source": "<a href='https://ads.twitter.com' rel='nofollow'>Twitter Ads</a>",
            "in_reply_to_status_id": null,
            "in_reply_to_status_id_str": null,
            "in_reply_to_user_id": null,
            "in_reply_to_user_id_str": null,
            "in_reply_to_screen_name": null,
            "user": {
                "id": 357750891,
                "id_str": "357750891",
                "name": "Twitter Marketing",
                "screen_name": "TwitterMktg",
                "location": "Twitter HQ ",
                "description": "Twitter’s place for marketers, agencies, and creative thinkers ⭐ Bringing you insights, news, updates, and inspiration. Visit @TwitterAdsHelp for Ads support.",
                "url": "https://t.co/Tfo4moo92y",
                "entities": {
                    "url": {
                        "urls": [
                            {
                                "url": "https://t.co/Tfo4moo92y",
                                "expanded_url": "https://marketing.twitter.com",
                                "display_url": "marketing.twitter.com",
                                "indices": [
                                    0,
                                    23
                                ]
                            }
                        ]
                    },
                    "description": {
                        "urls": [

                        ]
                    }
                },
                "protected": false,
                "followers_count": 924546,
                "friends_count": 661,
                "listed_count": 3893,
                "created_at": "Thu Aug 18 21:08:15 +0000 2011",
                "favourites_count": 1934,
                "utc_offset": -25200,
                "time_zone": "Pacific Time (US & Canada)",
                "geo_enabled": true,
                "verified": true,
                "statuses_count": 6329,
                "lang": "en",
                "contributors_enabled": false,
                "is_translator": false,
                "is_translation_enabled": false,
                "profile_background_color": "C0DEED",
                "profile_background_image_url": "http://pbs.twimg.com/profile_background_images/662767273/jvmxdpdrplhxcw8yvkv2.png",
                "profile_background_image_url_https": "https://pbs.twimg.com/profile_background_images/662767273/jvmxdpdrplhxcw8yvkv2.png",
                "profile_background_tile": true,
                "profile_image_url": "http://pbs.twimg.com/profile_images/800953549697888256/UlXXL5h5_normal.jpg",
                "profile_image_url_https": "https://pbs.twimg.com/profile_images/800953549697888256/UlXXL5h5_normal.jpg",
                "profile_banner_url": "https://pbs.twimg.com/profile_banners/357750891/1487188210",
                "profile_link_color": "19CF86",
                "profile_sidebar_border_color": "FFFFFF",
                "profile_sidebar_fill_color": "DDEEF6",
                "profile_text_color": "333333",
                "profile_use_background_image": true,
                "has_extended_profile": false,
                "default_profile": false,
                "default_profile_image": false,
                "following": false,
                "follow_request_sent": false,
                "notifications": false,
                "translator_type": "none"
            },
            "geo": null,
            "coordinates": null,
            "place": null,
            "contributors": null,
            "is_quote_status": false,
            "retweet_count": 111,
            "favorite_count": 162,
            "favorited": false,
            "retweeted": false,
            "possibly_sensitive": false,
            "lang": "en"
        },
        "is_quote_status": false,
        "retweet_count": 111,
        "favorite_count": 0,
        "favorited": false,
        "retweeted": false,
        "lang": "en"
    }
]"""

RSS = b"""
<?xml version="1.0" encoding="utf-8" ?>
<feed xmlns="http://www.w3.org/2005/Atom">
  <title>Pellentesque Tortor</title>
  <link href="https://signagewidgets.net/samples/rss.xml" />
  <entry>
    <id>6</id>
    <title>Fusce Ligula Egestas Vestibulum</title>
    <content>Lorem ipsum dolor sit amet, consectetur adipiscing elit. Donec sed odio dui egestas eget quam.</content>
    <link rel="enclosure" type="image/jpeg" href="https://signagewidgets.net/samples/1.jpg" />
    <published>2021-08-30T10:45:12Z</published>
  </entry>
  <entry>
    <id>1</id>
    <title>Ligula Amet Venenatis Consectetur</title>
    <content>Donec ullamcorper nulla non metus auctor fringilla. Vestibulum id ligula porta felis euismod semper.</content>
    <link rel="enclosure" type="image/jpeg" href="https://signagewidgets.net/samples/2.jpg" />
    <published>2021-08-30T14:17:51Z</published>
  </entry>
  <entry>
    <id>10</id>
    <title>Condimentum Purus Commodo, donec ullamcorper nulla non metus auctor fringilla</title>
    <content>Aenean lacinia bibendum nulla sed consectetur. Fusce dapibus, tellus ac cursus commodo, ut fermentum massa justo sit amet risus.</content>
    <link rel="enclosure" type="image/jpeg" href="https://signagewidgets.net/samples/3.jpg" />
    <published>2021-08-30T15:20:30Z</published>
  </entry>
  <entry>
    <id>4</id>
    <title>Parturient Nibh Mattis Vestibulum</title>
    <content>Cras justo odio, dapibus ac facilisis in, egestas eget quam. Integer posuere erat a ante venenatis dapibus posuere velit aliquet.</content>
    <published>2021-08-30T15:20:30Z</published>
  </entry>
  <entry>
    <id>5</id>
    <title>Ridiculus Tellus Magna</title>
    <published>2021-08-30T15:20:30Z</published>
  </entry>
  <entry>
    <id>7</id>
    <title>Fusce Ligula Egestas Vestibulum</title>
    <link rel="enclosure" type="image/jpeg" href="https://signagewidgets.net/samples/4.jpg" />
    <published>2021-08-30T10:45:12Z</published>
  </entry>
  <entry>
    <id>8</id>
    <content>Cras justo odio, dapibus ac facilisis in, egestas eget quam. Integer posuere erat a ante venenatis dapibus posuere velit aliquet.</content>
    <link rel="enclosure" type="image/jpeg" href="https://signagewidgets.net/samples/5.jpg" />
    <published>2021-08-30T10:45:12Z</published>
  </entry>
  <entry>
    <id>9</id>
    <link rel="enclosure" type="image/jpeg" href="https://signagewidgets.net/samples/6.jpg" />
    <published>2021-08-30T10:45:12Z</published>
  </entry>
</feed>
"""

VIDEOS = [
    {
        "type": "video",
        "url": "https://signagewidgets.net/samples/1.mp4",
        "duration": 48.693,
        "width": 720,
        "height": 720,
    },
    {
        "type": "video",
        "url": "https://signagewidgets.net/samples/2.mp4",
        "duration": 54.73,
        "width": 720,
        "height": 720,
    },
    {
        "type": "video",
        "url": "https://signagewidgets.net/samples/3.mp4",
        "duration": 17.067,
        "width": 640,
        "height": 640,
    },
    {
        "type": "video",
        "url": "https://signagewidgets.net/samples/4.mp4",
        "duration": 16.05,
        "width": 1920,
        "height": 1080,
    },
    {
        "type": "video",
        "url": "https://signagewidgets.net/samples/5.mp4",
        "duration": 15.723,
        "width": 1920,
        "height": 1080,
    },
    {
        "type": "video",
        "url": "https://signagewidgets.net/samples/6.mp4",
        "duration": 27.904,
        "width": 1920,
        "height": 1080,
    },
    {
        "type": "video",
        "url": "https://signagewidgets.net/samples/7.mp4",
        "duration": 55.126,
        "width": 1080,
        "height": 1920,
    },
    {
        "type": "video",
        "url": "https://signagewidgets.net/samples/8.mp4",
        "duration": 10.07,
        "width": 640,
        "height": 640,
    },
]



import colorsys
import hashlib
import logging
import re
import threading
from collections import OrderedDict
from html.parser import HTMLParser
from pathlib import Path

from jinja2 import nodes, Undefined, StrictUndefined, Markup
from jinja2.exceptions import TemplateSyntaxError
from jinja2.ext import Extension, GETTEXT_FUNCTIONS, extract_from_ast
from jinja2.sandbox import ImmutableSandboxedEnvironment

from . import utils


_local = threading.local()

logger = logging.getLogger("onsigntv.app_config")

BANNED_NAMES = {
    "category",
    "id",
    "list",
    "media",
    "name",
    "widget_id",
    "widget_uid",
    "gettext",
    "ngettext",
    "signagewidgets",
}

KNOWN_TYPES = {
    "audio",
    "audiolist",
    "airport",
    "bool",
    "choice",
    "color",
    "currency",
    "date",
    "datetime",
    "float",
    "font",
    "googlesheet",
    "googlevoice",
    "instagram",
    "image",
    "imagelist",
    "location",
    "media",
    "multichoice",
    "multicurrency",
    "number",
    "paragraph",
    "richtext",
    "stock",
    "text",
    "time",
    "twitter",
    "url",
    "video",
    "webfeed",
    "xml",
}

USER_TYPES = {"audio", "audiolist", "image", "imagelist", "media", "video"}

KNOWN_METAS = {"aspectratio", "caps", "compatibility", "description"}

DATA_SOURCE_TYPES = {
    "boolean",
    "date",
    "datetime",
    "image",
    "integer",
    "media",
    "number",
    "text",
    "time",
    "url",
    "video",
}

FETCH_TEMPLATE = """
<script type="text/plain" id="%(id)s">%(content)s</script>
<script type="text/javascript">
  window.%(variable)s = document.getElementById('%(id)s').innerHTML.trim().replace(/&(lt|gt|#34|#39|amp);/g, function(match) {
    switch(match) {
      case '&lt;': return '<'; case '&gt;': return '>'; case '&#34;': return '"'; case '&#39;': return "'"; case '&amp;': return '&';
    }
  });
</script>
"""


class MediaItem:
    def __init__(self, path, file_name, tracker):
        path = path / file_name

        if path.is_file():
            logger.debug(f"register media: {path}")
            self._path = path
            self._tracker = tracker
        else:
            raise ValueError(f"No file found for {path}")

    @property
    def name(self):
        return self._path.name

    @property
    def url(self):
        self._tracker(self._path)
        return "/" + str(self._path.relative_to(Path.cwd()))

    @property
    def width(self):
        from PIL import Image

        image = Image.open(self._path)
        return image.size[0]

    @property
    def height(self):
        from PIL import Image

        image = Image.open(self._path)
        return image.size[1]


class ConfigurableWidgetMedia:
    def __init__(self, path, tracker):
        logger.debug(f"register widget media: {path}")
        self._path = path
        self._tracker = tracker

    def __getitem__(self, name):
        return MediaItem(self._path, name, self._tracker)

    def __getattr__(self, key):
        for subpath in self._path.iterdir():
            if not subpath.name.startswith("."):
                if re.sub("[^\w]", "_", subpath.stem.lower()) == key:
                    return MediaItem(subpath, "", self._tracker)


class LocalContextManager:
    def __init__(self):
        self.ctx = {}
        self.provider = None
        self.manager = None

    def __call__(self, **kwargs):
        self.ctx = dict(kwargs)
        return self

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.ctx = {}

    def do_gettext(self, value):
        return value

    def do_ngettext(self, singular, plural, num):
        return singular if num == 1 else plural

    def do_localize(self, value, format=None):
        return value

    def do_shim(self, name):
        shims = {
            "events": "static/shim/events.js",
            "i18n": "static/shim/Intl.min.js",
        }

        path = shims.get(name)
        if path is None:
            raise ValueError("Unknown shim")

        if self.provider is None:
            from pkg_resources import ResourceManager, get_provider

            self.provider = get_provider("app_simulator")
            self.manager = ResourceManager()

        return Markup(
            self.provider.get_resource_string(self.manager, path).decode("utf-8")
        )

    def do_datasink(
        self, *, name, label, fields, help=None, optional=False, optgroup=None
    ):
        data_source_item = self.ctx["context"].get(name)
        if data_source_item:
            return data_source_item._render(name)
        return ""

    def do_fetch_feed(self, value):
        from .fields import WebFeed, WebFeedEntry
        from .rss import process_feed

        processed_feed = process_feed(value)

        return WebFeed(
            url=self.data,
            title=processed_feed["title"],
            subtitle=processed_feed["subtitle"],
            entries=[WebFeedEntry(entry) for entry in processed_feed["entries"]],
        )

    def do_fetch_sheet(self, value):
        from .fields import GoogleSheetsURLField, GoogleSheet

        match = GoogleSheetsURLField.RE_SHEET_ID.match(value)
        return GoogleSheet(match.groups(1))

    def do_fetch_text(self, url, refresh=15, variable="FETCH_RESULT"):
        try:
            resp = utils.get_url(url)
            data = resp.data
        except Exception:
            data = ""

        return Markup(FETCH_TEMPLATE) % {
            "id": "v" + hashlib.sha1(data).hexdigest(),
            "content": data.decode("utf-8"),
            "variable": variable,
        }

    def do_signagewidgets(self, *args, **kwargs):
        from urllib.parse import urlencode

        url = "https://signagewidgets.net/" + "/".join(map(str, args))

        if kwargs:
            url = url + "?" + urlencode(kwargs)

        return url


def render_app_html(context, path, tracker, **kwargs):
    context = {**context, "media": ConfigurableWidgetMedia(path.parent, tracker)}

    html = None
    exceptions = None

    env = default_jinja_env()

    try:

        # First we need to convert the data into a template
        template = env.from_string(path.read_text("utf-8"))

        with env.local_ctx(tracker=tracker, context=context):
            html = template.render(context)
    except Exception as errors:
        logger.debug("handling render exception")
        exceptions = errors

    return html, exceptions


def default_jinja_env():
    if hasattr(_local, "env"):
        return _local.env

    env = ImmutableSandboxedEnvironment(
        autoescape=True,
        keep_trailing_newline=True,
        undefined=StrictUndefined,
        extensions=[
            ErrorExtension,
            "jinja2.ext.autoescape",
            "jinja2.ext.InternationalizationExtension",
        ],
    )
    env.local_ctx = LocalContextManager()
    env.install_gettext_callables(
        env.local_ctx.do_gettext, env.local_ctx.do_ngettext, newstyle=True
    )

    # In the default env __config__ is a no-op.
    env.globals["__config__"] = lambda *args, **kwargs: ""
    env.globals["__meta__"] = lambda *args, **kwargs: ""

    env.globals["contrast"] = do_contrast
    env.globals["darken"] = do_darken
    env.globals["lighten"] = do_lighten
    env.globals["regex_match"] = do_regex_match
    env.globals["shim"] = env.local_ctx.do_shim
    env.globals["fetch_feed"] = env.local_ctx.do_fetch_feed
    env.globals["fetch_text"] = env.local_ctx.do_fetch_text
    env.globals["fetch_sheet"] = env.local_ctx.do_fetch_sheet
    env.globals["localize"] = env.local_ctx.do_localize
    env.globals["signagewidgets"] = env.local_ctx.do_signagewidgets
    env.globals["__datasink__"] = env.local_ctx.do_datasink
    env.globals["__datafeed__"] = env.local_ctx.do_datasink
    env.globals["__field__"] = lambda *args, **kwargs: ""

    env.globals["__lang__"] = "en"
    env.globals["__rtl__"] = False
    env.globals["__muted__"] = False
    env.globals["timeline_is_muted"] = False

    env.filters["numberfmt"] = do_numberfmt
    env.filters["slugify"] = do_slugify
    env.filters["qrcode"] = do_qrcode

    _local.env = env

    return env


def s_rgba(value):
    match = re.match(r"\s*#([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})\s*", value, re.I)
    if match:
        return tuple(int(i, 16) / 255 for i in match.groups()) + (1.0,)

    match = re.match(
        r"\s*#([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})\s*", value, re.I
    )
    if match:
        return tuple(int(i, 16) / 255 for i in match.groups())

    match = re.match(
        r"""
        rgba\(\s*
            ([01]?[0-9]?[0-9]|2[0-4][0-9]|25[0-5])\s*,\s*
            ([01]?[0-9]?[0-9]|2[0-4][0-9]|25[0-5])\s*,\s*
            ([01]?[0-9]?[0-9]|2[0-4][0-9]|25[0-5])\s*,\s*
            (\d*\.\d+)
        \s*\)
    """,
        value,
        re.X | re.I,
    )
    if match:
        return tuple(int(i) / 255 for i in match.groups()[:3]) + (
            float(match.groups()[-1]),
        )

    raise ValueError("invalid color")


def rgb_s(r, g, b, a=1.0):
    if a == 1.0:
        return "#%02x%02x%02x" % tuple(round(c * 255) for c in (r, g, b))
    else:
        return "rgba(%i,%i,%i,%.3f)" % tuple([round(c * 255) for c in (r, g, b)] + [a])


def s_amount(amount):
    if isinstance(amount, str):
        return float(amount.strip(" %")) / 100

    return amount


def do_lighten(value, amount=0.1):
    r, g, b, a = s_rgba(value)

    h, l, s = colorsys.rgb_to_hls(r, g, b)
    l = min(1, max(0, l + s_amount(amount)))

    return rgb_s(*colorsys.hls_to_rgb(h, l, s), a=a)


def do_darken(value, amount=0.1):
    r, g, b, a = s_rgba(value)

    h, l, s = colorsys.rgb_to_hls(r, g, b)
    l = min(1, max(0, l - s_amount(amount)))

    return rgb_s(*colorsys.hls_to_rgb(h, l, s), a=a)


def do_contrast(value, light, dark, threshold=0.43):
    r, g, b, a = s_rgba(value)

    # Gama correct the values
    r = r / 12.92 if r <= 0.03928 else ((r + 0.055) / 1.055) ** 2.4
    g = g / 12.92 if g <= 0.03928 else ((g + 0.055) / 1.055) ** 2.4
    b = b / 12.92 if b <= 0.03928 else ((b + 0.055) / 1.055) ** 2.4

    luma = 0.2126 * r + 0.7152 * g + 0.0722 * b

    if luma < threshold:
        return light
    else:
        return dark


def do_regex_match(pattern, string, case_insensitive=False):
    flags = 0
    if case_insensitive:
        flags = flags | re.I

    try:
        match = re.match(pattern, string, flags)
    except re.error:
        raise ValueError("Invalid regular expression")
    else:
        results = {}
        if match:
            results.update(
                dict(list(zip(list(range(len(match.groups()))), match.groups())))
            )
            results.update(match.groupdict())

        return results


def do_numberfmt(number, decimal_sep=".", thousand_sep=".", decimal_pos=2, grouping=3):
    """
    Adapted from https://github.com/django/django/blob/main/django/utils/numberformat.py
    Get a number (as a number or string), and return it as a string,
    using formats defined as arguments:

    * decimal_sep: Decimal separator symbol (for example ".")
    * thousand_sep: Thousand separator symbol (for example ",")
    * decimal_pos: Number of decimal positions
    * grouping: Number of digits in every group limited by thousand separator.
        For non-uniform digit grouping, it can be a sequence with the number
        of digit group sizes following the format used by the Python locale
        module in locale.localeconv() LC_NUMERIC grouping (e.g. (3, 2, 0)).
    """
    from decimal import Decimal

    # sign
    sign = ""
    if isinstance(number, Decimal):
        # Format values with more than 200 digits (an arbitrary cutoff) using
        # scientific notation to avoid high memory usage in {:f}'.format().
        _, digits, exponent = number.as_tuple()
        if abs(exponent) + len(digits) > 200:
            number = "{:e}".format(number)
            coefficient, exponent = number.split("e")
            # Format the coefficient.
            coefficient = do_numberfmt(
                coefficient,
                decimal_sep,
                thousand_sep,
                decimal_pos,
                grouping,
            )
            return "{}e{}".format(coefficient, exponent)
        else:
            str_number = "{:f}".format(number)
    else:
        str_number = str(number)
    if str_number[0] == "-":
        sign = "-"
        str_number = str_number[1:]

    # decimal part
    if "." in str_number:
        int_part, dec_part = str_number.split(".")
        if decimal_pos is not None:
            dec_part = dec_part[:decimal_pos]
    else:
        int_part, dec_part = str_number, ""
    if decimal_pos is not None:
        dec_part = dec_part + ("0" * (decimal_pos - len(dec_part)))
    dec_part = dec_part and decimal_sep + dec_part

    # grouping
    try:
        # if grouping is a sequence
        intervals = list(grouping)
    except TypeError:
        # grouping is a single value
        intervals = [grouping, 0]
    active_interval = intervals.pop(0)
    int_part_gd = ""
    cnt = 0
    for digit in int_part[::-1]:
        if cnt and cnt == active_interval:
            if intervals:
                active_interval = intervals.pop(0) or active_interval
            int_part_gd += thousand_sep[::-1]
            cnt = 0
        int_part_gd += digit
        cnt += 1
    int_part = int_part_gd[::-1]

    return sign + int_part + dec_part


def do_slugify(value):
    import unicodedata

    value = str(value)
    value = (
        unicodedata.normalize("NFKD", value).encode("ascii", "ignore").decode("ascii")
    )
    value = re.sub(r"[^\w\s-]", "", value).strip().lower()
    return re.sub(r"[-\s]+", "-", value)


def do_qrcode(value):
    import qrcode
    import qrcode.image.svg

    from io import BytesIO

    qr_svg = qrcode.make(value, image_factory=qrcode.image.svg.SvgPathImage)

    svg_stream = BytesIO()
    qr_svg.save(svg_stream)

    svg_value = svg_stream.getvalue().decode("utf-8")

    return Markup(svg_value[svg_value.find("<svg") :])


class LenientUndefined(Undefined):
    """Allows the user to do pretty much anything with undefined values."""

    def _dont_give_up(self, *args, **kwargs):
        return LenientUndefined()

    def __getattr__(self, name):
        if name[:2] == "__":
            raise AttributeError(name)
        return self._dont_give_up()

    def __int__(self):
        return 0

    def __float__(self):
        return 0.0

    __add__ = (
        __radd__
    ) = (
        __mul__
    ) = (
        __rmul__
    ) = (
        __div__
    ) = (
        __rdiv__
    ) = (
        __truediv__
    ) = (
        __rtruediv__
    ) = (
        __floordiv__
    ) = (
        __rfloordiv__
    ) = (
        __mod__
    ) = (
        __rmod__
    ) = (
        __pos__
    ) = (
        __neg__
    ) = (
        __call__
    ) = (
        __getitem__
    ) = (
        __lt__
    ) = __le__ = __gt__ = __ge__ = __complex__ = __pow__ = __rpow__ = _dont_give_up


class ErrorExtension(Extension):
    tags = set(["error"])

    def parse(self, parser):
        lineno = next(parser.stream).lineno

        return nodes.Output(
            [self.call_method("_raise", [parser.parse_expression()])]
        ).set_lineno(lineno)

    def _raise(self, message):
        if self.environment.undefined is LenientUndefined:
            return ""

        raise ValueError(message)


def lenient_jinja_env():
    if hasattr(_local, "lenient"):
        return _local.lenient

    lenient = ImmutableSandboxedEnvironment(
        autoescape=True,
        keep_trailing_newline=True,
        undefined=LenientUndefined,
        extensions=[
            ErrorExtension,
            "jinja2.ext.autoescape",
            "jinja2.ext.InternationalizationExtension",
        ],
    )
    lenient.filters = LenientUndefined()
    lenient.install_null_translations(newstyle=True)

    _local.lenient = lenient

    return lenient


class HTMLTemplateParser(HTMLParser):
    def __init__(self):
        self.convert_charrefs = True
        self.reset()
        self._in_title = False
        self.title = None
        self.variables = []
        self.metas = {}

    def handle_starttag(self, tag, attrs):
        attrs = dict(attrs)
        self._in_title = tag == "title"

        if tag == "meta" and all([attrs.get(a) for a in ("label", "name", "type")]):
            self.variables.append((self.getpos(), attrs))
        elif (
            tag == "meta" and attrs.get("content") and attrs.get("name") in KNOWN_METAS
        ):
            self.metas[attrs["name"]] = attrs["content"]

    def handle_endtag(self, tag):
        self._in_title = False

    def handle_data(self, data):
        if self._in_title:
            self.title = data.strip()


def extract_app_config(template_text):
    config = {}
    env = lenient_jinja_env()

    template = env.from_string(template_text)

    fields = OrderedDict()
    extra_fields = OrderedDict()
    exceptions = []

    def detect_config(
        name,
        type,
        label,
        value=None,
        help=None,
        optional=False,
        *,
        optgroup=None,
        **kwargs,
    ):
        field = {
            "type": type,
            "label": label,
            "value": value,
            "help_text": help,
            "required": not bool(optional),
            "optgroup": optgroup,
        }

        if type in ("choice", "multichoice"):
            if not isinstance(kwargs.get("choices"), (list, tuple)):
                raise ValueError(
                    'A "choices" argument is required and must be a list: %(name)s'
                    % {"name": name}
                )

            for choice in kwargs["choices"]:
                if not isinstance(choice, (list, tuple)):
                    raise ValueError(
                        'Each item in "choices" must be a list: %(name)s'
                        % {"name": name}
                    )

                if len(choice) != 2:
                    raise ValueError(
                        'Each item in "choices" must have two strings (name, label): %(name)s'
                        % {"name": name}
                    )

                if not (isinstance(choice[0], str) and isinstance(choice[1], str)):
                    raise ValueError(
                        'Each item in "choices" must have two strings (name, label): %(name)s'
                        % {"name": name}
                    )

            if value:
                if value not in [choice[0] for choice in kwargs["choices"]]:
                    raise ValueError(
                        'Default value must be present in "choices" argument: %(name)s'
                        % {"name": name}
                    )
            else:
                field["value"] = kwargs["choices"][0][0]

            field["choices"] = kwargs["choices"]

        fields[name] = field

    class DataSinkField:
        def __init__(self, *, name, type, label, help=None, optional=False):
            if not re.match("^[a-zA-Z][a-zA-Z0-9_]*$", name):
                raise ValueError("Invalid field name: %(name)s" % {"name": name})

            if type not in DATA_SOURCE_TYPES:
                raise ValueError("Data source type invalid: %(type)s" % {"type": type})

            self.name = name
            self.type = type
            self.label = label
            self.help_text = help
            self.required = not bool(optional)

        def __repr__(self):
            return "<DataSinkField {}>".format(self.as_dict())

        def as_dict(self):
            return self.__dict__

    def detect_datasink(
        *, name, label, fields, help=None, optional=False, optgroup=None
    ):
        if not re.match("^[a-zA-Z][a-zA-Z0-9_]*$", name):
            raise ValueError("Invalid data sink name: %(name)s" % {"name": name})

        if not isinstance(fields, list):
            raise ValueError("Data sink fields must be a list")

        if not fields:
            raise ValueError("Data sink fields must not be empty")

        if not all([isinstance(f, DataSinkField) for f in fields]):
            raise ValueError("Data sink fields must use __field__ constructor")

        field_datasink = {
            "type": "datasink",
            "label": label,
            "value": None,
            "help_text": help,
            "required": not bool(optional),
            "datasource_fields": [f.as_dict() for f in fields],
            "optgroup": optgroup,
        }

        extra_fields[name] = field_datasink

    def detect_meta(name, value):
        if name in KNOWN_METAS:
            config[name] = value

    # Then we render the template to remove all Jinja2 tags
    html = template.render(
        {
            "__config__": detect_config,
            "__meta__": detect_meta,
            "__datasink__": detect_datasink,
            "__datafeed__": detect_datasink,
            "__field__": DataSinkField,
        }
    )

    # Now we use a lenient parser to try not to gobble up on invalid markup
    parser = HTMLTemplateParser()
    parser.feed(html)

    if parser.title:
        config["title"] = parser.title
    else:
        raise ValueError("A title tag is required for this app")

    config.update(parser.metas)

    if parser.variables:
        if fields:
            raise ValueError("Unable to mix __config__ and <meta> variables together.")

        for pos, attrs in parser.variables:
            name = attrs["name"]
            field = {
                "type": attrs["type"],
                "label": attrs["label"],
                "value": attrs.get("value"),
                "help_text": attrs.get("help"),
                "required": "optional" not in attrs,
                "optgroup": attrs.get("optgroup") or None,
            }
            if field["value"] == "":
                field["value"] = None

            if field["type"] in ("choice", "multichoice"):
                if not field.get("value"):
                    raise ValueError(
                        "Choice variable requires a value: %(name)s" % {"name": name}
                    )

                if name not in fields:
                    fields[name] = field
                    fields[name]["choices"] = [[field["value"], field["label"]]]
                elif fields[name]["type"] not in ("choice", "multichoice"):
                    raise ValueError(
                        "Choice variable configured wrong: %(name)s" % {"name": name}
                    )
                else:
                    fields[name]["choices"].append([field["value"], field["label"]])
            else:
                fields[name] = field

        for field in [
            f[1] for f in fields.items() if f[1]["type"] in ("choice", "multichoice")
        ]:
            choices = field["choices"]

            field["label"] = ""
            if all([re.match(r".*\S.*:.*\S.*", choice[1]) for choice in choices]):
                field["label"] = choices[0][1].split(":", 1)[0].strip()

                for choice in choices:
                    choice[1] = choice[1].split(":", 1)[1].strip()

    for name, field in fields.items():
        if not re.match("^[a-zA-Z][a-zA-Z0-9_]*$", name):
            raise ValueError(
                {"error": "Invalid variable name: %(name)s" % {"name": name}}
            )

        if name in env.globals or name in BANNED_NAMES:
            raise ValueError(
                {"error": "Invalid variable name: %(name)s" % {"name": name}}
            )

        if field["type"] not in KNOWN_TYPES:
            raise ValueError(
                {"error": "Invalid variable type: %(name)s" % {"name": name}}
            )

        if field["type"] in USER_TYPES and field.get("value"):
            raise ValueError(
                "User variables cannot contain default values: %(name)s"
                % {"name": name}
            )

    config["fields"] = list(map(list, fields.items()))
    config["exceptions"] = exceptions

    try:
        node = env.parse(template_text)
        list(env.lex(env.preprocess(template_text)))
    except TemplateSyntaxError:
        pass
    else:
        catalog = OrderedDict()

        for lineno, func, message in extract_from_ast(
            node, GETTEXT_FUNCTIONS, babel_style=False
        ):

            if message not in catalog:
                catalog[message] = {"msgid": message}

        if catalog:
            config["i18n"] = True
            config["catalog"] = list(catalog.values())

    if extra_fields:
        config["fields"] = [] if not config.get("fields") else config["fields"]
        config["fields"] += [
            [field_name, field_data] for field_name, field_data in extra_fields.items()
        ]

    return config
